#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include "contrast.h"
#include "edges.h"

int Num_Orientations = 4;
double Sigma_c = 1.3;
double Sigma_s = 1.3;
int O_Contrast = 1;
int I_Contrast = 0;
int O_Target = 180;

main(int argc, char *argv[])
{

  unsigned char *edge_data[4], *char_con_map;
  double **gaus_c, **gaus_s;
  double **con_map, *norm_con_map;
  int level_size, max;
  double max2;
  int i, j, k;
  int Min_Radius = 3;
  int Max_Radius = 3;
  char imname[30];
  FILE *data_file;
  int orien;


  gaus_c=(double **) calloc(2*Max_Radius+1, sizeof(double*));
  for (i=0; i<2*Max_Radius+1; i++)
    gaus_c[i]=(double *) calloc(2*Max_Radius+1, sizeof(double));

  gaus_s=(double **) calloc(2*(3*Max_Radius+1)+1, sizeof(double*));
  for (i=0; i<2*(3*Max_Radius+1)+1; i++)
    gaus_s[i]=(double *) calloc(2*(3*Max_Radius+1)+1, sizeof(double));

  max = read_pgm(argv[1], &edge_data[0], &level_size, &level_size);
  max = read_pgm(argv[2], &edge_data[1], &level_size, &level_size);
  max = read_pgm(argv[3], &edge_data[2], &level_size, &level_size);
  max = read_pgm(argv[4], &edge_data[3], &level_size, &level_size);

  norm_con_map = (double *) malloc(level_size*level_size*sizeof(double));
  char_con_map = (unsigned char *) malloc(level_size*level_size);

  for (i=Min_Radius; i<=Max_Radius; i+=2)
  {
    max=contrast(edge_data,
                 &con_map,
                 gaus_c, gaus_s,
                 level_size, i, 4);

    for (orien=1; orien<=Num_Orientations; orien++)
    {
      max2 = 0;
      for (k=0; k<level_size; k++)
        for (j=0; j<level_size; j++)
  	  if(con_map[orien][k*level_size+j]>max2) 
	    max2=con_map[orien][k*level_size+j];
      printf("max2 = %g\n", max2);
    
      /* normalize contrast map */
      norm_dbl(con_map[orien], norm_con_map, level_size*level_size, max2);

      /* write out double image data  
      data_file = fopen("norm_dbl_data", "w");
      for (k=0; k<level_size; k++)
      {
      	for (j=0; j<level_size; j++)
       	  fprintf(data_file, "%d ", (int)norm_con_map[k*level_size+j]);
        fprintf(data_file, "\n");
      }
      */

      /* convert to unsigned char */
      d_to_c(norm_con_map, char_con_map, level_size*level_size);

      /* make image name */
      if(orien==1) sprintf(imname, "gaus045.%d.pgm", i);
      if(orien==2) sprintf(imname, "gaus090.%d.pgm", i);

      /* write as pgm image */
      write_pgm(imname, char_con_map, level_size, level_size, 255);

    }
  }
}

void norm_dbl(double *from, double *to, int size, double max)
{
  int i;
  double factor=255.0/max;
  FILE *fa, *fb;

  for(i=0; i<size; i++) 
  {
    to[i] = from[i]>0.0 ? from[i]*factor : 0.0;
  }

}

void d_to_c(double *from, unsigned char *to, int size)
{
  int i;
  for (i=0; i<size; i++)
    to[i] = from[i]>0.0 ? (unsigned char)from[i] :0;
}


double contrast(unsigned char **feature_map,  
		double **(*dbl_contrast),
		double **gaus_c, double **gaus_s,
		int width, 
		int radius_c, 
		int num_features)
{
  int size = width * width;
  int radius_s=radius_c*3+1;
  unsigned char *char_map;
  double max_inten;
  int i, j;
  double max;

  static int level=0;
  char f_name[20];
  unsigned char *temp_c;
  double *temp_d;

  FILE *f_ptr;

  (*dbl_contrast) = (double **) calloc(num_features,  sizeof(double *));
  for (i=0; i<num_features; i++)
  {
    (*dbl_contrast)[i] = (double *) calloc(size,  sizeof(double));
  }
  
  /* calculate values for gaussians */
  Calc_Gaus(gaus_c, radius_c, radius_c*Sigma_c);
  Calc_Gaus(gaus_s, radius_s, radius_s*Sigma_s);

  f_ptr = fopen("gaus_c", "w");
  for(i=0; i<2*radius_c+1; i++)
  {
    for (j=0; j<2*radius_c+1; j++)
      fprintf(f_ptr, "%g\t", gaus_c[i][j]);
    fprintf(f_ptr, "\n");
  }
  fclose(f_ptr);

  f_ptr = fopen("gaus_s", "w");
  for(i=0; i<2*radius_s+1; i++)
  {
    for (j=0; j<2*radius_s+1; j++)
      fprintf(f_ptr, "%g\t", gaus_s[i][j]);
    fprintf(f_ptr, "\n");
  }
  fclose(f_ptr);

  max=Feature_Contrast(width, radius_c, radius_s, gaus_c, gaus_s,
			feature_map, (*dbl_contrast), num_features);

  return max;

} /* contrast()*/


void combine_maps(double max, double *values, double ***o_contrast,
		  double ***i_contrast, int index, int size)
{
  int i, j;
  double tmax, factor;
  char f_name[20];
  double **contrast_feature;
  int start, end;
  int o_start=0, o_end;

  o_end = Num_Orientations;
  factor=255.0/max;

  printf("max=%g\n", max);

  if(!(O_Contrast || I_Contrast))
  {
    fprintf(stderr, "ERROR: No contrast dimension\n");
    return;
  }

  if (O_Target < 180)
  {
    o_start = (float)(O_Target)/180 * Num_Orientations;
    o_end = o_start+1;
  }

  if(O_Contrast && I_Contrast)
  {
    for (i=0; i<size; i++)
    {
      tmax=0;

      for (j=0; j<2; j++)
        if (i_contrast[index][j][i] > tmax)
  	  tmax=i_contrast[index][j][i];

      for (j=o_start; j<o_end; j++)
        if (o_contrast[index][j][i] > tmax)
  	  tmax=o_contrast[index][j][i];

      values[i] = (double) (tmax*factor);
    } /* for */
  } /* if */
  else 
  {
    if(O_Contrast) 
    {
      contrast_feature=o_contrast[index];
      start = o_start;
      end = o_end;
    }
    else 
    {
      contrast_feature=i_contrast[index];
      start = 0;
      end = 2;
    }

    printf("start = %d, end = %d\n", start, end);
    for (i=0; i<size; i++)
    {
      tmax=0;

      for (j=start; j<end; j++)
        if (contrast_feature[j][i] > tmax)
  	  tmax=contrast_feature[j][i];

      values[i] = (double) (tmax*factor);
    } /* for */
  } /* else */
} /* combine_maps */


double Feature_Contrast(int width, int center_radius, int radius, 
		   double **gaus_c, double **gaus_s,
		   unsigned char **edges, double **contrast_map,
		   int num_features)
{
  int i, j, pixel_x, pixel_y;
  int temp;
  float weight[num_features];
  double cell[num_features];
  double max=0.0;

  Calc_Weights(weight, num_features);

  for (pixel_y=radius; pixel_y<width-radius; pixel_y++)
    for (pixel_x=radius; pixel_x<width-radius; pixel_x++)
    {
      for(i=0; i<num_features; i++)
      {
        cell[i]=WEIGHT*Convolve(pixel_x, pixel_y, width,
                         center_radius, gaus_c, edges[i]);
        cell[i]=cell[i]-Convolve(pixel_x, pixel_y, width, radius,
                                 gaus_s, edges[i]);
/* remove comments to display DOG results */
	contrast_map[i][pixel_y*width+pixel_x]=cell[i];
      }

      for(i=0; i<num_features; i++)
      {
/* insert comments to display DOG results 
        for(j=0; j<num_features; j++)
	{
          contrast_map[i][pixel_y*width+pixel_x] 
			+= weight[(j-i+num_features)%num_features]
			   * cell[j];
	}
/* comments end here */
        if (contrast_map[i][pixel_y*width+pixel_x]>max)
	  max=contrast_map[i][pixel_y*width+pixel_x];
      } /* for (i) */

    }/* for (pixels) */

  return max;

}/* Feature_Contrast()*/


void Calc_Gaus(double **gaus, int rad, double sigma)
{
  int i,j;
  double total=0.0;

  for(i=-rad; i<=rad; i++)
    for (j=-rad; j<=rad; j++)
    {
      gaus[i+rad][j+rad]=4*exp(-(float)(i*i+j*j)/(2.0*sigma));
      total+=gaus[i+rad][j+rad];
    }

  for(i=-rad; i<=rad; i++)
    for (j=-rad; j<=rad; j++)
      gaus[i+rad][j+rad]=gaus[i+rad][j+rad]/total;
}/* Calc_Gaus() */

void Calc_Weights(float *w, int num_features)
{
  int i;
  float denom;
 
  w[0]=1.0;
  w[1]=-0.15;
  w[2]=-0.8;
  w[3]=-0.15;

/*
  w[0]=1.0;
  w[1]=1.0;

  denom=num_features*num_features;
  if (num_features%2)
    denom=denom-1;
 

  for (i=1; i<=num_features/2; i++)
  {
    w[i]=(float)(-4*i)/denom;
    w[num_features-i]=w[i];
  }
*/
} /* Calc_Weights() */


double Convolve(int x, int y, int width, int rad, double **gaus,
                unsigned char *values)
{
  int i,j;
  double temp=0;

  for(i=-rad; i<=rad; i++)
    for(j=-rad; j<=rad; j++)
        temp+=(double) values[(y-i)*width+(x-j)] *
                gaus[i+rad][j+rad];

  return temp;
} /* Convolve */














